@@ -15,7 +15,7 @@ template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
 template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
 #endif
 template <class S, class T> static inline void clone(T*& dst, S* src, int n)
-{   
+{
 	dst = new T[n];
 	memcpy((void *)dst,(void *)src,sizeof(T)*n);
 }
@@ -378,12 +378,12 @@ double l2r_l2_svr_fun::fun(double *w)
 	int l=prob->l;
 	int w_size=get_nr_variable();
 	double d;
-	
+
 	Xv(w, z);
 
 	for(i=0;i<w_size;i++)
 		f += w[i]*w[i];
-	f /= 2;	
+	f /= 2;
 	for(i=0;i<l;i++)
 	{
 		d = z[i] - y[i];
@@ -408,7 +408,7 @@ void l2r_l2_svr_fun::grad(double *w, double *g)
 	for(i=0;i<l;i++)
 	{
 		d = z[i] - y[i];
-		
+
 		// generate index set I
 		if(d < -p)
 		{
@@ -422,7 +422,7 @@ void l2r_l2_svr_fun::grad(double *w, double *g)
 			I[sizeI] = i;
 			sizeI++;
 		}
-	
+
 	}
 	subXTv(z, g);
 
@@ -559,7 +559,7 @@ double Solver_MCSVM_CS::Solve(double *w)
 		alpha[i] = 0;
 
 	for(i=0;i<w_size*nr_class;i++)
-		w[i] = 0; 
+		w[i] = 0;
 	for(i=0;i<l;i++)
 	{
 		for(m=0;m<nr_class;m++)
@@ -570,7 +570,7 @@ double Solver_MCSVM_CS::Solve(double *w)
 		{
 			double val = xi->value;
 			QD[i] += val*val;
-	
+
 			// Uncomment the for loop if initial alpha isn't zero
 			// for(m=0; m<nr_class; m++)
 			//	w[(xi->index-1)*nr_class+m] += alpha[i*nr_class+m]*val;
@@ -581,7 +581,7 @@ double Solver_MCSVM_CS::Solve(double *w)
 		index[i] = i;
 	}
 
-	while(iter < max_iter) 
+	while(iter < max_iter)
 	{
 		double stopping = -INF;
 		for(i=0;i<active_size;i++)
@@ -632,14 +632,14 @@ double Solver_MCSVM_CS::Solve(double *w)
 						active_size_i[i]--;
 						while(active_size_i[i]>m)
 						{
-							if(!be_shrunk(i, active_size_i[i], y_index[i], 
+							if(!be_shrunk(i, active_size_i[i], y_index[i],
 											alpha_i[alpha_index_i[active_size_i[i]]], minG))
 							{
 								swap(alpha_index_i[m], alpha_index_i[active_size_i[i]]);
 								swap(G[m], G[active_size_i[i]]);
 								if(y_index[i] == active_size_i[i])
 									y_index[i] = m;
-								else if(y_index[i] == m) 
+								else if(y_index[i] == m)
 									y_index[i] = active_size_i[i];
 								break;
 							}
@@ -652,7 +652,7 @@ double Solver_MCSVM_CS::Solve(double *w)
 				{
 					active_size--;
 					swap(index[s], index[active_size]);
-					s--;	
+					s--;
 					continue;
 				}
 
@@ -778,7 +778,7 @@ double Solver_MCSVM_CS::Solve(double *w)
 // To support weights for instances, use GETI(i) (i)
 
 static double solve_l2r_l1l2_svc(
-	const problem *prob, double *w, double eps, 
+	const problem *prob, double *w, double eps,
 	double Cp, double Cn, int solver_type)
 {
 	int l = prob->l;
@@ -813,7 +813,7 @@ static double solve_l2r_l1l2_svc(
 	{
 		if(prob->y[i] > 0)
 		{
-			y[i] = +1; 
+			y[i] = +1;
 		}
 		else
 		{
@@ -825,7 +825,7 @@ static double solve_l2r_l1l2_svc(
 	// 0 <= alpha[i] <= upper_bound[GETI(i)]
 	for(i=0; i<l; i++)
 		alpha[i] = 0;
-	
+
 	for(i=0; i<w_size; i++)
 		w[i] = 0;
 	for(i=0; i<l; i++)
@@ -1233,11 +1233,11 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 	int i, s, iter = 0;
 	double *xTx = new double[l];
 	int max_iter = 1000;
-	int *index = new int[l];		
+	int *index = new int[l];	
 	double *alpha = new double[2*l]; // store alpha and C - alpha
-	schar *y = new schar[l];	
+	schar *y = new schar[l];
 	int max_inner_iter = 100; // for inner Newton
-	double innereps = 1e-2; 
+	double innereps = 1e-2;
 	double innereps_min = min(1e-8, eps);
 	double upper_bound[3] = {Cn, 0, Cp};
 
@@ -1245,14 +1245,14 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 	{
 		if(prob->y[i] > 0)
 		{
-			y[i] = +1; 
+			y[i] = +1;
 		}
 		else
 		{
 			y[i] = -1;
 		}
 	}
-		
+	
 	// Initial alpha can be set here. Note that
 	// 0 < alpha[i] < upper_bound[GETI(i)]
 	// alpha[2*i] + alpha[2*i+1] = upper_bound[GETI(i)]
@@ -1304,7 +1304,7 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 
 			// Decide to minimize g_1(z) or g_2(z)
 			int ind1 = 2*i, ind2 = 2*i+1, sign = 1;
-			if(0.5*a*(alpha[ind2]-alpha[ind1])+b < 0) 
+			if(0.5*a*(alpha[ind2]-alpha[ind1])+b < 0)
 			{
 				ind1 = 2*i+1;
 				ind2 = 2*i;
@@ -1314,7 +1314,7 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 			//  g_t(z) = z*log(z) + (C-z)*log(C-z) + 0.5a(z-alpha_old)^2 + sign*b(z-alpha_old)
 			double alpha_old = alpha[ind1];
 			double z = alpha_old;
-			if(C - z < 0.5 * C) 
+			if(C - z < 0.5 * C)
 				z = 0.1*z;
 			double gp = a*(z-alpha_old)+sign*b+log(z/(C-z));
 			Gmax = max(Gmax, fabs(gp));
@@ -1322,13 +1322,13 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 			// Newton method on the sub-problem
 			const double eta = 0.1; // xi in the paper
 			int inner_iter = 0;
-			while (inner_iter <= max_inner_iter) 
+			while (inner_iter <= max_inner_iter)
 			{
 				if(fabs(gp) < innereps)
 					break;
 				double gpp = a + C/(C-z)/z;
 				double tmpz = z - gp/gpp;
-				if(tmpz <= 0) 
+				if(tmpz <= 0)
 					z *= eta;
 				else // tmpz in (0, C)
 					z = tmpz;
@@ -1346,7 +1346,7 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 				{
 					w[xi->index-1] += sign*(z-alpha_old)*yi*xi->value;
 					xi++;
-				}  
+				}
 			}
 		}
 
@@ -1354,10 +1354,10 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 		if(iter % 10 == 0)
 			info(".");
 
-		if(Gmax < eps) 
+		if(Gmax < eps)
 			break;
 
-		if(newton_iter <= l/10) 
+		if(newton_iter <= l/10)
 			innereps = max(innereps_min, 0.1*innereps);
 
 	}
@@ -1367,13 +1367,13 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 		info("\nWARNING: reaching max number of iterations\nUsing -s 0 may be faster (also see FAQ)\n\n");
 
 	// calculate objective value
-	
+
 	double v = 0;
 	for(i=0; i<w_size; i++)
 		v += w[i] * w[i];
 	v *= 0.5;
 	for(i=0; i<l; i++)
-		v += alpha[2*i] * log(alpha[2*i]) + alpha[2*i+1] * log(alpha[2*i+1]) 
+		v += alpha[2*i] * log(alpha[2*i]) + alpha[2*i+1] * log(alpha[2*i+1])
 			- upper_bound[GETI(i)] * log(upper_bound[GETI(i)]);
 	info("Objective value = %lf\n", v);
 
@@ -1381,7 +1381,7 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 	delete [] alpha;
 	delete [] y;
 	delete [] index;
-	
+
 	return v;
 }
 
@@ -1403,7 +1403,7 @@ double solve_l2r_lr_dual(const problem *prob, double *w, double eps, double Cp,
 // To support weights for instances, use GETI(i) (i)
 
 static double solve_l1r_l2_svc(
-	problem *prob_col, double *w, double eps, 
+	problem *prob_col, double *w, double eps,
 	double Cp, double Cn)
 {
 	int l = prob_col->l;
@@ -1670,7 +1670,7 @@ static double solve_l1r_l2_svc(
 	delete [] y;
 	delete [] b;
 	delete [] xj_sq;
-	
+
 	return v;
 }
 
@@ -1692,7 +1692,7 @@ static double solve_l1r_l2_svc(
 // To support weights for instances, use GETI(i) (i)
 
 static double solve_l1r_lr(
-	const problem *prob_col, double *w, double eps, 
+	const problem *prob_col, double *w, double eps,
 	double Cp, double Cn)
 {
 	int l = prob_col->l;
@@ -2024,7 +2024,7 @@ static double solve_l1r_lr(
 		info("WARNING: reaching max number of iterations\n");
 
 	// calculate objective value
-	
+
 	double v = 0;
 	int nnz = 0;
 	for(j=0; j<w_size; j++)
@@ -2063,8 +2063,8 @@ static void transpose(const problem *prob, feature_node **x_space_ret, problem *
 	int i;
 	int l = prob->l;
 	int n = prob->n;
-	int nnz = 0;
-	int *col_ptr = new int[n+1];
+	long int nnz = 0;
+	long int *col_ptr = new long int [n+1];
 	feature_node *x_space;
 	prob_col->l = l;
 	prob_col->n = n;
@@ -2152,6 +2152,24 @@ static void group_classes(const problem *prob, int *nr_class_ret, int **label_re
 		}
 	}
 
+	//
+	// Labels are ordered by their first occurrence in the training set. 
+	// However, for two-class sets with -1/+1 labels and -1 appears first, 
+	// we swap labels to ensure that internally the binary SVM has positive data corresponding to the +1 instances.
+	//
+	if (nr_class == 2 && label[0] == -1 && label[1] == 1)
+	{
+		swap(label[0],label[1]);
+		swap(count[0],count[1]);
+		for(i=0;i<l;i++)
+		{
+			if(data_label[i] == 0)
+				data_label[i] = 1;
+			else
+				data_label[i] = 0;
+		}
+	}
+
 	int *start = Malloc(int,nr_class);
 	start[0] = 0;
 	for(i=1;i<nr_class;i++)
@@ -2183,7 +2201,7 @@ static double train_one(const problem *prob, const parameter *param, double *w,
 	neg = prob->l - pos;
 
 	double objective = 0.0;
-	
+
 	double primal_solver_tol = eps*max(min(pos,neg), 1)/prob->l;
 
 	function *fun_obj=NULL;
@@ -2204,7 +2222,7 @@ static double train_one(const problem *prob, const parameter *param, double *w,
 			tron_obj.set_print_string(liblinear_print_string);
 			objective = tron_obj.tron(w);
 			delete fun_obj;
-			delete C;
+			delete[] C;
 			break;
 		}
 		case L2R_L2LOSS_SVC:
@@ -2222,7 +2240,7 @@ static double train_one(const problem *prob, const parameter *param, double *w,
 			tron_obj.set_print_string(liblinear_print_string);
 			objective = tron_obj.tron(w);
 			delete fun_obj;
-			delete C;
+			delete[] C;
 			break;
 		}
 		case L2R_L2LOSS_SVC_DUAL:
@@ -2261,13 +2279,13 @@ static double train_one(const problem *prob, const parameter *param, double *w,
 			double *C = new double[prob->l];
 			for(int i = 0; i < prob->l; i++)
 				C[i] = param->C;
-			
+
 			fun_obj=new l2r_l2_svr_fun(prob, C, param->p);
 			TRON tron_obj(fun_obj, param->eps);
 			tron_obj.set_print_string(liblinear_print_string);
 			objective = tron_obj.tron(w);
 			delete fun_obj;
-			delete C;
+			delete[] C;
 			break;
 
 		}
@@ -2385,6 +2403,7 @@ model* train(const problem *prob, const parameter *param)
 			}
 			else
 			{
+				model_->objective = 0;
 				model_->w=Malloc(double, w_size*nr_class);
 				double *w=Malloc(double, w_size);
 				for(i=0;i<nr_class;i++)
@@ -2400,7 +2419,7 @@ model* train(const problem *prob, const parameter *param)
 					for(; k<sub_prob.l; k++)
 						sub_prob.y[k] = -1;
 
-					model_->objective = train_one(&sub_prob, param, w, weighted_C[i], param->C);
+					model_->objective += train_one(&sub_prob, param, w, weighted_C[i], param->C) / nr_class;
 
 					for(int j=0;j<w_size;j++)
 						model_->w[j*nr_class+i] = w[j];
@@ -2425,10 +2444,15 @@ model* train(const problem *prob, const parameter *param)
 void cross_validation(const problem *prob, const parameter *param, int nr_fold, double *target)
 {
 	int i;
-	int *fold_start = Malloc(int,nr_fold+1);
+	int *fold_start;
 	int l = prob->l;
 	int *perm = Malloc(int,l);
-
+	if (nr_fold > l)
+	{
+		nr_fold = l;
+		fprintf(stderr,"WARNING: # folds > # data. Will use # folds = # data instead (i.e., leave-one-out cross validation)\n");
+	}
+	fold_start = Malloc(int,nr_fold+1);
 	for(i=0;i<l;i++) perm[i]=i;
 	for(i=0;i<l;i++)
 	{
@@ -2560,7 +2584,7 @@ double predict_probability(const struct model *model_, const struct feature_node
 				prob_estimates[i]=prob_estimates[i]/sum;
 		}
 
-		return label;		
+		return label;
 	}
 	else
 		return 0;
@@ -2600,7 +2624,7 @@ int save_model(const char *model_file_name, const struct model *model_)
 
 	fprintf(fp, "solver_type %s\n", solver_type_table[param.solver_type]);
 	fprintf(fp, "nr_class %d\n", model_->nr_class);
-	
+
 	if(model_->label)
 	{
 		fprintf(fp, "label");
@@ -2643,7 +2667,7 @@ struct model *load_model(const char *model_file_name)
 	parameter& param = model_->param;
 
 	model_->label = NULL;
-	
+
 	char *old_locale = strdup(setlocale(LC_ALL, NULL));
 	setlocale(LC_ALL, "C");
 
@@ -2666,7 +2690,7 @@ struct model *load_model(const char *model_file_name)
 			if(solver_type_table[i] == NULL)
 			{
 				fprintf(stderr,"unknown solver type.\n");
-				
+
 				setlocale(LC_ALL, old_locale);
 				free(model_->label);
 				free(model_);
@@ -2731,10 +2755,10 @@ struct model *load_model(const char *model_file_name)
 			fscanf(fp, "%lf ", &model_->w[i*nr_w+j]);
 		fscanf(fp, "\n");
 	}
-	
+
 	setlocale(LC_ALL, old_locale);
 	free(old_locale);
-	
+
 	if (ferror(fp) != 0 || fclose(fp) != 0) return NULL;
 
 	return model_;
@@ -2790,7 +2814,7 @@ const char *check_parameter(const problem *prob, const parameter *param)
 
 	if(param->C <= 0)
 		return "C <= 0";
-	
+
 	if(param->p < 0)
 		return "p < 0";
 
@@ -2819,7 +2843,7 @@ int check_probability_model(const struct model *model_)
 
 void set_print_string_function(void (*print_func)(const char*))
 {
-	if (print_func == NULL) 
+	if (print_func == NULL)
 		liblinear_print_string = &print_string_stdout;
 	else
 		liblinear_print_string = print_func;
