@@ -2,6 +2,8 @@
 //  Copyright(C) 2010-2013 Taro Watanabe <taro.watanabe@nict.go.jp>
 //
 
+#include <queue>
+
 #include "cicada_alignment_impl.hpp"
 
 #include "utils/resource.hpp"
@@ -9,6 +11,8 @@
 #include "utils/compress_stream.hpp"
 #include "utils/lockfree_list_queue.hpp"
 #include "utils/bithack.hpp"
+#include "utils/double_base64_parser.hpp"
+#include "utils/double_base64_generator.hpp"
 
 #include <boost/program_options.hpp>
 #include <boost/filesystem.hpp>
@@ -706,27 +710,22 @@ struct ViterbiMapReduce
   
   struct bitext_type
   {
-    bitext_type() : id(size_type(-1)), source(), target(), alignment() {}
+    bitext_type() : id(size_type(-1)), source(), target() {}
     bitext_type(const size_type& __id, const sentence_type& __source, const sentence_type& __target)
-      : id(__id), source(__source), target(__target), alignment() {}
-    bitext_type(const size_type& __id, const sentence_type& __source, const sentence_type& __target, const alignment_type& __alignment)
-      : id(__id), source(__source), target(__target), alignment(__alignment) {}
+      : id(__id), source(__source), target(__target) {}
     
     bitext_type(const size_type& __id,
 		const sentence_type& __source, const sentence_type& __target,
-		const span_set_type& __span_source, const span_set_type& __span_target,
-		const alignment_type& __alignment)
+		const span_set_type& __span_source, const span_set_type& __span_target)
       : id(__id),
 	source(__source), target(__target),
-	span_source(__span_source), span_target(__span_target),
-	alignment(__alignment) {}
+	span_source(__span_source), span_target(__span_target) {}
     
     size_type     id;
     sentence_type source;
     sentence_type target;
     span_set_type span_source;
     span_set_type span_target;
-    alignment_type alignment;
 
     void clear()
     {
@@ -735,7 +734,6 @@ struct ViterbiMapReduce
       target.clear();
       span_source.clear();
       span_target.clear();
-      alignment.clear();
     }
     
     void swap(bitext_type& x)
@@ -745,11 +743,57 @@ struct ViterbiMapReduce
       target.swap(x.target);
       span_source.swap(x.span_source);
       span_target.swap(x.span_target);
-      alignment.swap(x.alignment);
+    }
+
+    friend
+    bool operator<(const bitext_type& x, const bitext_type& y)
+    {
+      return x.id < y.id;
+    }
+
+    friend
+    bool operator>(const bitext_type& x, const bitext_type& y)
+    {
+      return x.id > y.id;
+    }
+  };
+
+  struct viterbi_type
+  {
+    size_type id;
+    std::string output;
+    
+    viterbi_type() : id(size_type(-1)), output() {}
+    
+    void clear()
+    {
+      id = size_type(-1);
+      output.clear();
+    }
+
+    void swap(viterbi_type& x)
+    {
+      std::swap(id, x.id);
+      output.swap(x.output);
+    }
+
+    friend
+    bool operator<(const viterbi_type& x, const viterbi_type& y)
+    {
+      return x.id < y.id;
+    }
+    
+    friend
+    bool operator>(const viterbi_type& x, const viterbi_type& y)
+    {
+      return x.id > y.id;
     }
   };
   
-  typedef utils::lockfree_list_queue<bitext_type, std::allocator<bitext_type> > queue_type;
+  typedef utils::lockfree_list_queue<bitext_type, std::allocator<bitext_type> > queue_mapper_type;
+  typedef utils::lockfree_list_queue<viterbi_type, std::allocator<viterbi_type> > queue_reducer_type;
+  
+  typedef std::vector<queue_reducer_type, std::allocator<queue_reducer_type> > queue_reducer_set_type;
 };
 
 namespace std
@@ -759,16 +803,25 @@ namespace std
   {
     x.swap(y);
   }
+
+  inline
+  void swap(ViterbiMapReduce::viterbi_type& x, ViterbiMapReduce::viterbi_type& y)
+  {
+    x.swap(y);
+  }
 };
 
 template <typename Aligner>
 struct ViterbiMapper : public ViterbiMapReduce, public Aligner
 {
-  queue_type& mapper;
-  queue_type& reducer_source_target;
-  queue_type& reducer_target_source;
+  queue_mapper_type& mapper;
+  queue_reducer_type& reducer_source_target;
+  queue_reducer_type& reducer_target_source;
   
-  ViterbiMapper(const Aligner& __aligner, queue_type& __mapper, queue_type& __reducer_source_target, queue_type& __reducer_target_source)
+  ViterbiMapper(const Aligner& __aligner,
+		queue_mapper_type& __mapper,
+		queue_reducer_type& __reducer_source_target,
+		queue_reducer_type& __reducer_target_source)
     : Aligner(__aligner),
       mapper(__mapper),
       reducer_source_target(__reducer_source_target),
@@ -778,9 +831,13 @@ struct ViterbiMapper : public ViterbiMapReduce, public Aligner
   void operator()()
   {
     bitext_type bitext;
+    
     alignment_type alignment_source_target;
     alignment_type alignment_target_source;
-
+    
+    viterbi_type viterbi_source_target;
+    viterbi_type viterbi_target_source;
+    
     const int iter_mask = (1 << 12) - 1;
     
     for (int iter = 0; /**/; ++ iter) {
@@ -792,78 +849,83 @@ struct ViterbiMapper : public ViterbiMapReduce, public Aligner
       
       if (! bitext.source.empty() && ! bitext.target.empty())
 	Aligner::operator()(bitext.source, bitext.target, bitext.span_source, bitext.span_target, alignment_source_target, alignment_target_source);
+
+      viterbi_source_target.id = bitext.id;
+      viterbi_target_source.id = bitext.id;
       
-      reducer_source_target.push(bitext_type(bitext.id, bitext.source, bitext.target, alignment_source_target));
-      reducer_target_source.push(bitext_type(bitext.id, bitext.target, bitext.source, alignment_target_source));
+      write(viterbi_source_target, bitext.id, bitext.source, bitext.target, alignment_source_target);
+      write(viterbi_target_source, bitext.id, bitext.target, bitext.source, alignment_target_source);
+      
+      reducer_source_target.push_swap(viterbi_source_target);
+      reducer_target_source.push_swap(viterbi_target_source);
 
       if ((iter & iter_mask) == iter_mask)
 	Aligner::shrink();
     }
+    
+    reducer_source_target.push(viterbi_type());
+    reducer_target_source.push(viterbi_type());
   }
-};
-
-struct ViterbiReducer : public ViterbiMapReduce
-{
-  struct less_bitext
-  {
-    bool operator()(const bitext_type& x, const bitext_type& y) const
-    {
-      return x.id < y.id;
-    }
-  };
-  typedef std::set<bitext_type, less_bitext, std::allocator<bitext_type> > bitext_set_type;
-
-  typedef boost::shared_ptr<std::ostream> ostream_ptr_type;
   
-  ostream_ptr_type os;
-  queue_type& queue;
-  
-  ViterbiReducer(const path_type& path, queue_type& __queue) : os(), queue(__queue) 
-  {
-    if (! path.empty()) {
-      const bool flush_output = (path == "-"
-				 || (boost::filesystem::exists(path)
-				     && ! boost::filesystem::is_regular_file(path)));
-      
-      os.reset(new utils::compress_ostream(path, 1024 * 1024 * (! flush_output)));
-    }
-  }
-
   typedef int index_type;
   typedef std::vector<index_type, std::allocator<index_type> > index_set_type;
   typedef std::vector<index_set_type, std::allocator<index_set_type> > align_set_type;
   typedef std::set<index_type, std::less<index_type>, std::allocator<index_type> > align_none_type;
+
+  typedef std::vector<char, std::allocator<char> > buffer_type;
   
   align_set_type  aligns;
   align_none_type aligns_none;
-  
-  void write(std::ostream& os, const bitext_type& bitext)
+
+  buffer_type buffer;
+
+  void write(viterbi_type& viterbi,
+	     const size_type& id,
+	     const sentence_type& source,
+	     const sentence_type& target,
+	     const alignment_type& alignment)
   {
-    if (moses_mode)
-      os << bitext.alignment << '\n';
-    else {
-      os << "# Sentence pair (" << (bitext.id + 1) << ')'
-	 << " source length " << bitext.source.size()
-	 << " target length " << bitext.target.size()
-	 << " alignment score : " << 0 << '\n';
-      os << bitext.target << '\n';
+    typedef std::ostream_iterator<char> oiter_type;
     
-      if (bitext.alignment.empty() || bitext.source.empty() || bitext.target.empty()) {
-	os << "NULL ({ })";
-	sentence_type::const_iterator siter_end = bitext.source.end();
-	for (sentence_type::const_iterator siter = bitext.source.begin(); siter != siter_end; ++ siter)
-	  os << ' ' << *siter << " ({ })";
+    namespace karma = boost::spirit::karma;
+    namespace standard = boost::spirit::standard;
+
+    buffer.clear();
+    
+    boost::iostreams::filtering_ostream os;
+    os.push(boost::iostreams::back_inserter(buffer));
+    
+    if (moses_mode)
+      os << alignment << '\n';
+    else {
+      os << "# Sentence pair (" << (id + 1) << ')'
+	 << " source length " << source.size()
+	 << " target length " << target.size()
+	 << " alignment score : " << 0 << '\n';
+      os << target << '\n';
+    
+      if (source.empty() || target.empty() || alignment.empty()) {
+	os << "NULL";
+	os << " ({";
+	for (size_type trg = 0; trg != target.size(); ++ trg)
+	  os << ' ' << (trg + 1);
+	os << " })";
+	
+	for (size_type src = 0; src != source.size(); ++ src) {
+	  os << ' ' << source[src];
+	  os << " ({ })";
+	}
 	os << '\n';
       } else {
 	aligns.clear();
-	aligns.resize(bitext.source.size());
+	aligns.resize(source.size());
       
 	aligns_none.clear();
-	for (size_type trg = 0; trg != bitext.target.size(); ++ trg)
+	for (size_type trg = 0; trg != target.size(); ++ trg)
 	  aligns_none.insert(trg + 1);
       
-	alignment_type::const_iterator aiter_end = bitext.alignment.end();
-	for (alignment_type::const_iterator aiter = bitext.alignment.begin(); aiter != aiter_end; ++ aiter) {
+	alignment_type::const_iterator aiter_end = alignment.end();
+	for (alignment_type::const_iterator aiter = alignment.begin(); aiter != aiter_end; ++ aiter) {
 	  aligns[aiter->source].push_back(aiter->target + 1);
 	  aligns_none.erase(aiter->target + 1);
 	}
@@ -873,8 +935,8 @@ struct ViterbiReducer : public ViterbiMapReduce
 	std::copy(aligns_none.begin(), aligns_none.end(), std::ostream_iterator<index_type>(os, " "));
 	os << "})";
       
-	for (size_type src = 0; src != bitext.source.size(); ++ src) {
-	  os << ' ' << bitext.source[src];
+	for (size_type src = 0; src != source.size(); ++ src) {
+	  os << ' ' << source[src];
 	  os << " ({ ";
 	  std::copy(aligns[src].begin(), aligns[src].end(), std::ostream_iterator<index_type>(os, " "));
 	  os << "})";
@@ -882,46 +944,124 @@ struct ViterbiReducer : public ViterbiMapReduce
 	os << '\n';
       }
     }
+    
+    os.reset();
+    
+    viterbi.output = std::string(buffer.begin(), buffer.end());
+  }
+};
+
+struct ViterbiReducer : public ViterbiMapReduce
+{
+  typedef std::pair<viterbi_type, queue_reducer_type*> viterbi_queue_type;
+  typedef std::vector<viterbi_queue_type, std::allocator<viterbi_queue_type> > viterbi_queue_set_type;
+
+  struct heap_compare_type
+  {
+    bool operator()(const viterbi_queue_type* x, const viterbi_queue_type* y) const
+    {
+      return x->first > y->first;
+    }
+  };
+  
+  typedef std::vector<viterbi_queue_type*, std::allocator<viterbi_queue_type*> > heap_base_type;
+  typedef std::priority_queue<viterbi_queue_type*, heap_base_type, heap_compare_type> heap_type;
+  
+  typedef std::vector<viterbi_queue_type*, std::allocator<viterbi_queue_type*> > queue_type;
+  
+  typedef boost::shared_ptr<std::ostream> ostream_ptr_type;
+  
+  ostream_ptr_type os;
+  queue_reducer_set_type& queues;
+  bool flush_;
+  
+  ViterbiReducer(const path_type& path, queue_reducer_set_type& __queues) : os(), queues(__queues), flush_(false)
+  {
+    if (! path.empty()) {
+      flush_ = (path == "-"
+		|| (boost::filesystem::exists(path)
+		    && ! boost::filesystem::is_regular_file(path)));
+      
+      os.reset(new utils::compress_ostream(path, 1024 * 1024));
+    }
   }
   
-  void operator()()
+  void operator()() throw()
   {
-    if (! os ) {
-      bitext_type bitext;
-      for (;;) {
-	queue.pop_swap(bitext);
-	if (bitext.id == size_type(-1)) break;
-      }
-    } else {
-      bitext_set_type bitexts;
-      size_type id = 0;
-      bitext_type bitext;
-      for (;;) {
-	queue.pop_swap(bitext);
-	if (bitext.id == size_type(-1)) break;
-	
-	if (bitext.id == id) {
-	  write(*os, bitext);
-	  ++ id;
-	} else
-	  bitexts.insert(bitext);
-	
-	while (! bitexts.empty() && bitexts.begin()->id == id) {
-	  write(*os, *bitexts.begin());
-	  bitexts.erase(bitexts.begin());
-	  ++ id;
-	}
-      }
+    std::ostream* stream = os.get();
+
+    viterbi_queue_set_type viterbis(queues.size());
+    
+    heap_type heap;
+    queue_type queue;
+    queue_type queue_next;
+
+    size_type id = 0;
+    
+    for (size_type shard = 0; shard != queues.size(); ++ shard) {
+      viterbis[shard].second = &queues[shard];
       
-      while (! bitexts.empty() && bitexts.begin()->id == id) {
-	write(*os, *bitexts.begin());
-	bitexts.erase(bitexts.begin());
-	++ id;
-      }
-      
-      if (! bitexts.empty())
-	throw std::runtime_error("error while writeing viterbi output?");
+      queue.push_back(&viterbis[shard]);
     }
+    
+    int non_found_iter = 0;
+    while (! heap.empty() || ! queue.empty()) {
+      bool found = false;
+
+      if (! heap.empty() && heap.top()->first.id == id) {
+	viterbi_queue_type* viterbi_queue = heap.top();
+	heap.pop();
+	
+	++ id;
+	
+	if (stream) {
+	  *stream << viterbi_queue->first.output;
+	  
+	  if (flush_)
+	    *stream << std::flush;
+	}
+	
+	queue.push_back(viterbi_queue);
+	
+	found = true;
+      }
+
+      if (! queue.empty()) {
+	for (size_type i = 0; i != queue.size(); ++ i) {
+	  if (queue[i]->second->pop_swap(queue[i]->first, true)) {
+	    if (queue[i]->first.id != size_type(-1))
+	      heap.push(queue[i]);
+	    
+	    found = true;
+	  } else
+	    queue_next.push_back(queue[i]);
+	}
+	
+	queue.swap(queue_next);
+	queue_next.clear();
+      }
+      
+      non_found_iter = loop_sleep(found, non_found_iter);
+    }
+  }
+
+  int loop_sleep(bool found, int non_found_iter)
+  {
+    if (! found) {
+      boost::thread::yield();
+      ++ non_found_iter;
+    } else
+      non_found_iter = 0;
+  
+    if (non_found_iter >= 50) {
+      struct timespec tm;
+      tm.tv_sec = 0;
+      tm.tv_nsec = 2000001;
+      nanosleep(&tm, NULL);
+    
+      non_found_iter = 0;
+    }
+    return non_found_iter;
   }
 };
 
@@ -932,14 +1072,17 @@ void viterbi(const ttable_type& ttable_source_target,
   typedef ViterbiReducer         reducer_type;
   typedef ViterbiMapper<Aligner> mapper_type;
   
-  typedef reducer_type::bitext_type bitext_type;
-  typedef reducer_type::queue_type  queue_type;
+  typedef reducer_type::bitext_type    bitext_type;
+
+  typedef reducer_type::queue_mapper_type      queue_mapper_type;
+  typedef reducer_type::queue_reducer_type     queue_reducer_type;
+  typedef reducer_type::queue_reducer_set_type queue_reducer_set_type;
   
   typedef std::vector<mapper_type, std::allocator<mapper_type> > mapper_set_type;
 
-  queue_type queue(threads * 4096);
-  queue_type queue_source_target;
-  queue_type queue_target_source;
+  queue_mapper_type     queue(threads * 1024);
+  queue_reducer_set_type queue_source_target(threads, queue_reducer_type(1024));
+  queue_reducer_set_type queue_target_source(threads, queue_reducer_type(1024));
 
   boost::thread_group reducer;
   reducer.add_thread(new boost::thread(reducer_type(viterbi_source_target_file, queue_source_target)));
@@ -949,8 +1092,8 @@ void viterbi(const ttable_type& ttable_source_target,
   for (int i = 0; i != threads; ++ i)
     mapper.add_thread(new boost::thread(mapper_type(Aligner(ttable_source_target, ttable_target_source),
 						    queue,
-						    queue_source_target,
-						    queue_target_source)));
+						    queue_source_target[i],
+						    queue_target_source[i])));
   
   bitext_type bitext;
   bitext.id = 0;
@@ -1005,12 +1148,6 @@ void viterbi(const ttable_type& ttable_source_target,
   }
 
   mapper.join_all();
-
-  bitext.clear();
-  queue_source_target.push_swap(bitext);
-  bitext.clear();
-  queue_target_source.push_swap(bitext);
-  
   reducer.join_all();
 
   utils::resource viterbi_end;
@@ -1053,30 +1190,43 @@ struct ProjectionMapReduce
       dependency_target.swap(x.dependency_target);
     }
   };
-  
+
   struct projected_type
   {
     size_type id;
-    dependency_type dependency;
+    std::string output;
     
-    projected_type() : id(size_type(-1)), dependency() {}
+    projected_type() : id(size_type(-1)), output() {}
     
     void clear()
     {
       id = size_type(-1);
-      dependency.clear();
+      output.clear();
     }
     
     void swap(projected_type& x)
     {
       std::swap(id, x.id);
-      dependency.swap(x.dependency);
+      output.swap(x.output);
+    }
+    
+    friend
+    bool operator<(const projected_type& x, const projected_type& y)
+    {
+      return x.id < y.id;
+    }
+    
+    friend
+    bool operator>(const projected_type& x, const projected_type& y)
+    {
+      return x.id > y.id;
     }
   };
   
-
   typedef utils::lockfree_list_queue<bitext_type, std::allocator<bitext_type> >       queue_mapper_type;
   typedef utils::lockfree_list_queue<projected_type, std::allocator<projected_type> > queue_reducer_type;
+  
+  typedef std::vector<queue_reducer_type, std::allocator<queue_reducer_type> > queue_reducer_set_type;
 };
 
 namespace std
@@ -1113,6 +1263,10 @@ struct ProjectionMapper : public ProjectionMapReduce, public Analyzer
   void operator()()
   {
     bitext_type bitext;
+
+    dependency_type dependency_source;
+    dependency_type dependency_target;
+    
     projected_type projected_source;
     projected_type projected_target;
     
@@ -1122,95 +1276,163 @@ struct ProjectionMapper : public ProjectionMapReduce, public Analyzer
       mapper.pop_swap(bitext);
       if (bitext.id == size_type(-1)) break;
 
-      projected_source.clear();
-      projected_target.clear();
+      dependency_source.clear();
+      dependency_target.clear();
       
       if (! bitext.source.empty() && ! bitext.target.empty())
 	Analyzer::operator()(bitext.source,
 			     bitext.target,
 			     bitext.dependency_source,
 			     bitext.dependency_target,
-			     projected_source.dependency,
-			     projected_target.dependency);
+			     dependency_source,
+			     dependency_target);
       
       projected_source.id = bitext.id;
       projected_target.id = bitext.id;
       
+      write(projected_source, dependency_source);
+      write(projected_target, dependency_target);
+      
       reducer_source.push_swap(projected_source);
       reducer_target.push_swap(projected_target);
       
       if ((iter & iter_mask) == iter_mask)
 	Analyzer::shrink();
     }
+    
+    reducer_source.push(projected_type());
+    reducer_target.push(projected_type());
+  }
+
+  typedef std::vector<char, std::allocator<char> > buffer_type;
+  
+  buffer_type buffer;
+  
+  void write(projected_type& projected, const dependency_type& dependency)
+  {
+    buffer.clear();
+    
+    boost::iostreams::filtering_ostream os;
+    os.push(boost::iostreams::back_inserter(buffer));
+    
+    os << dependency << '\n';
+    
+    projected.output = std::string(buffer.begin(), buffer.end());
   }
   
 };
 
 struct ProjectionReducer : public ProjectionMapReduce
 {
-  struct less_projected
+  typedef std::pair<projected_type, queue_reducer_type*> projected_queue_type;
+  typedef std::vector<projected_queue_type, std::allocator<projected_queue_type> > projected_queue_set_type;
+
+  struct heap_compare_type
   {
-    bool operator()(const projected_type& x, const projected_type& y) const
+    bool operator()(const projected_queue_type* x, const projected_queue_type* y) const
     {
-      return x.id < y.id;
+      return x->first > y->first;
     }
   };
-  typedef std::set<projected_type, less_projected, std::allocator<projected_type> > projected_set_type;
+  
+  typedef std::vector<projected_queue_type*, std::allocator<projected_queue_type*> > heap_base_type;
+  typedef std::priority_queue<projected_queue_type*, heap_base_type, heap_compare_type> heap_type;
+
+  typedef std::vector<projected_queue_type*, std::allocator<projected_queue_type*> > queue_type;
   
   typedef boost::shared_ptr<std::ostream> ostream_ptr_type;
   
   ostream_ptr_type    os;
-  queue_reducer_type& queue;
+  queue_reducer_set_type& queues;
+  bool flush_;
   
-  ProjectionReducer(const path_type& path, queue_reducer_type& __queue) : os(), queue(__queue)
+  ProjectionReducer(const path_type& path, queue_reducer_set_type& __queues) : os(), queues(__queues), flush_(false)
   {
     if (! path.empty()) {
-      const bool flush_output = (path == "-"
-				 || (boost::filesystem::exists(path)
-				     && ! boost::filesystem::is_regular_file(path)));
+      flush_ = (path == "-"
+		|| (boost::filesystem::exists(path)
+		    && ! boost::filesystem::is_regular_file(path)));
       
-      os.reset(new utils::compress_ostream(path, 1024 * 1024 * (! flush_output)));
+      os.reset(new utils::compress_ostream(path, 1024 * 1024));
     }
   }
   
-  void operator()()
+  void operator()() throw()
   {
-    if (! os) {
-      projected_type projected;
-      for (;;) {
-	queue.pop_swap(projected);
-	if (projected.id == size_type(-1)) break;
-      }
-    } else { 
-      size_type id = 0;
-      projected_type     projected;
-      projected_set_type buffer;
-      for (;;) {
-	queue.pop_swap(projected);
-	if (projected.id == size_type(-1)) break;
-       
-	if (projected.id == id) {
-	  *os << projected.dependency << '\n';
-	  ++ id;
-	} else
-	  buffer.insert(projected);
-       
-	while (! buffer.empty() && buffer.begin()->id == id) {
-	  *os << buffer.begin()->dependency << '\n';
-	  buffer.erase(buffer.begin());
-	  ++ id;
-	}
-      }
-     
-      while (! buffer.empty() && buffer.begin()->id == id) {
-	*os << buffer.begin()->dependency << '\n';
-	buffer.erase(buffer.begin());
-	++ id;
-      }
-     
-      if (! buffer.empty())
-	throw std::runtime_error("error while writing dependency output?");
+    std::ostream* stream = os.get();
+
+    projected_queue_set_type projections(queues.size());
+    
+    heap_type heap;
+    queue_type queue;
+    queue_type queue_next;
+
+    size_type id = 0;
+    
+    for (size_type shard = 0; shard != queues.size(); ++ shard) {
+      projections[shard].second = &queues[shard];
+      
+      queue.push_back(&projections[shard]);
     }
+    
+    int non_found_iter = 0;
+    while (! heap.empty() || ! queue.empty()) {
+      bool found = false;
+
+      if (! heap.empty() && heap.top()->first.id == id) {
+	projected_queue_type* projected_queue = heap.top();
+	heap.pop();
+	
+	++ id;
+	
+	if (stream) {
+	  *stream << projected_queue->first.output;
+	  
+	  if (flush_)
+	    *stream << std::flush;
+	}
+	
+	queue.push_back(projected_queue);
+	
+	found = true;
+      }
+
+      if (! queue.empty()) {
+	for (size_type i = 0; i != queue.size(); ++ i) {
+	  if (queue[i]->second->pop_swap(queue[i]->first, true)) {
+	    if (queue[i]->first.id != size_type(-1))
+	      heap.push(queue[i]);
+	    
+	    found = true;
+	  } else
+	    queue_next.push_back(queue[i]);
+	}
+	
+	queue.swap(queue_next);
+	queue_next.clear();
+      }
+      
+      non_found_iter = loop_sleep(found, non_found_iter);
+    }
+  }
+  
+  int loop_sleep(bool found, int non_found_iter)
+  {
+    if (! found) {
+      boost::thread::yield();
+      ++ non_found_iter;
+    } else
+      non_found_iter = 0;
+  
+    if (non_found_iter >= 50) {
+      struct timespec tm;
+      tm.tv_sec = 0;
+      tm.tv_nsec = 2000001;
+      nanosleep(&tm, NULL);
+    
+      non_found_iter = 0;
+    }
+    return non_found_iter;
   }
 };
 
@@ -1223,16 +1445,16 @@ void project_dependency(const ttable_type& ttable_source_target,
   
   typedef reducer_type::bitext_type    bitext_type;
   typedef reducer_type::projected_type projected_type;
-  
-  typedef reducer_type::queue_mapper_type  queue_mapper_type;
-  typedef reducer_type::queue_reducer_type queue_reducer_type;
+
+  typedef reducer_type::queue_mapper_type      queue_mapper_type;
+  typedef reducer_type::queue_reducer_type     queue_reducer_type;
+  typedef reducer_type::queue_reducer_set_type queue_reducer_set_type;
 
   typedef std::vector<mapper_type, std::allocator<mapper_type> > mapper_set_type;
-  
-  
-  queue_mapper_type  queue(threads * 4096);
-  queue_reducer_type queue_source;
-  queue_reducer_type queue_target;
+
+  queue_mapper_type  queue(threads * 1024);
+  queue_reducer_set_type queue_source(threads, queue_reducer_type(1024));
+  queue_reducer_set_type queue_target(threads, queue_reducer_type(1024));
   
   boost::thread_group reducer;
   reducer.add_thread(new boost::thread(reducer_type(projected_source_file, queue_source)));
@@ -1242,8 +1464,8 @@ void project_dependency(const ttable_type& ttable_source_target,
   for (int i = 0; i != threads; ++ i)
     mapper.add_thread(new boost::thread(mapper_type(Analyzer(ttable_source_target, ttable_target_source),
 						    queue,
-						    queue_source,
-						    queue_target)));
+						    queue_source[i],
+						    queue_target[i])));
     
   bitext_type bitext;
   bitext.id = 0;
@@ -1297,9 +1519,6 @@ void project_dependency(const ttable_type& ttable_source_target,
     queue.push_swap(bitext);
   }
   mapper.join_all();
-  
-  queue_source.push(projected_type());
-  queue_target.push(projected_type());
   reducer.join_all();
 
   utils::resource projection_end;
@@ -1342,25 +1561,39 @@ struct PosteriorMapReduce
   struct posterior_type
   {
     size_type id;
-    matrix_type matrix;
+    std::string output;
     
-    posterior_type() : id(size_type(-1)), matrix() {}
+    posterior_type() : id(size_type(-1)), output() {}
     
     void clear()
     {
       id = size_type(-1);
-      matrix.clear();
+      output.clear();
     }
-    
+
     void swap(posterior_type& x)
     {
       std::swap(id, x.id);
-      matrix.swap(x.matrix);
+      output.swap(x.output);
+    }
+
+    friend
+    bool operator<(const posterior_type& x, const posterior_type& y)
+    {
+      return x.id < y.id;
+    }
+    
+    friend
+    bool operator>(const posterior_type& x, const posterior_type& y)
+    {
+      return x.id > y.id;
     }
   };
   
   typedef utils::lockfree_list_queue<bitext_type, std::allocator<bitext_type> >       queue_mapper_type;
   typedef utils::lockfree_list_queue<posterior_type, std::allocator<posterior_type> > queue_reducer_type;
+  
+  typedef std::vector<queue_reducer_type, std::allocator<queue_reducer_type> > queue_reducer_set_type;
 };
 
 namespace std
@@ -1398,6 +1631,10 @@ struct PosteriorMapper : public PosteriorMapReduce, public Infer
   void operator()()
   {
     bitext_type    bitext;
+    
+    matrix_type matrix_source_target;
+    matrix_type matrix_target_source;
+    
     posterior_type posterior_source_target;
     posterior_type posterior_target_source;
     
@@ -1407,148 +1644,183 @@ struct PosteriorMapper : public PosteriorMapReduce, public Infer
       mapper.pop_swap(bitext);
       if (bitext.id == size_type(-1)) break;
       
-      posterior_source_target.clear();
-      posterior_target_source.clear();
+      matrix_source_target.clear();
+      matrix_target_source.clear();
       
       if (! bitext.source.empty() && ! bitext.target.empty())
-	Infer::operator()(bitext.source, bitext.target, posterior_source_target.matrix, posterior_target_source.matrix);
+	Infer::operator()(bitext.source, bitext.target, matrix_source_target, matrix_target_source);
       
       posterior_source_target.id = bitext.id;
       posterior_target_source.id = bitext.id;
       
+      write(posterior_source_target, matrix_source_target);
+      write(posterior_target_source, matrix_target_source);
+      
       reducer_source_target.push_swap(posterior_source_target);
       reducer_target_source.push_swap(posterior_target_source);
       
       if ((iter & iter_mask) == iter_mask)
 	Infer::shrink();
     }
-  }
-};
-
-struct PosteriorReducer : public PosteriorMapReduce
-{
-  struct less_posterior
-  {
-    bool operator()(const posterior_type& x, const posterior_type& y) const
-    {
-      return x.id < y.id;
-    }
-  };
-  typedef std::set<posterior_type, less_posterior, std::allocator<posterior_type> > posterior_set_type;
-  
-  typedef boost::shared_ptr<std::ostream> ostream_ptr_type;
-  
-  ostream_ptr_type os;
-  queue_reducer_type& queue;
-  
-  PosteriorReducer(const path_type& path, queue_reducer_type& __queue) : os(), queue(__queue)
-  {
-    if (! path.empty()) {
-      const bool flush_output = (path == "-"
-				 || (boost::filesystem::exists(path)
-				     && ! boost::filesystem::is_regular_file(path)));
-      
-      os.reset(new utils::compress_ostream(path, 1024 * 1024 * (! flush_output)));
-      os->precision(20);
-    }
+    
+    reducer_source_target.push(posterior_type());
+    reducer_target_source.push(posterior_type());
   }
   
-  void operator()()
+  struct real_precision : boost::spirit::karma::real_policies<long double>
   {
-    if (! os) {
-      posterior_type posterior;
-      for (;;) {
-	queue.pop_swap(posterior);
-	if (posterior.id == size_type(-1)) break;
-      }
-    } else {
-      posterior_set_type posteriors;
-      size_type      id = 0;
-      posterior_type posterior;
-      
-      for (;;) {
-	queue.pop_swap(posterior);
-	if (posterior.id == size_type(-1)) break;
-	
-	if (posterior.id == id) {
-	  write(*os, posterior);
-	  ++ id;
-	} else
-	  posteriors.insert(posterior);
-	
-	while (! posteriors.empty() && posteriors.begin()->id == id) {
-	  write(*os, *posteriors.begin());
-	  posteriors.erase(posteriors.begin());
-	  ++ id;
-	}
-      }
-      
-      while (! posteriors.empty() && posteriors.begin()->id == id) {
-	write(*os, *posteriors.begin());
-	posteriors.erase(posteriors.begin());
-	++ id;
-      }
-      
-      if (! posteriors.empty())
-	throw std::runtime_error("error while writeing posterior output?");
-    }
-  }
-
-  struct real_precision : boost::spirit::karma::real_policies<double>
-  {
-    static unsigned int precision(double) 
+    static unsigned int precision(long double) 
     { 
       return 20;
     }
   };
+
+  typedef std::vector<char, std::allocator<char> > buffer_type;
+
+  buffer_type buffer;
   
-  void write(std::ostream& os, const posterior_type& posterior)
+  void write(posterior_type& posterior, const matrix_type& matrix)
   {
-#if 0
     namespace karma = boost::spirit::karma;
     namespace standard = boost::spirit::standard;
     
-    typedef std::ostream_iterator<char> iterator_type;
+    typedef std::back_insert_iterator<buffer_type> iterator_type;
     
-    const matrix_type& matrix = posterior.matrix;
+    //karma::real_generator<long double, real_precision> real;
+    utils::double_base64_generator<iterator_type> base64;
 
-    karma::real_generator<double, real_precision> real;
-    iterator_type iter(os);
+    buffer.clear();
+    iterator_type iter(buffer);
 
     if (! matrix.empty()) {
       karma::generate(iter, karma::lit('('));
       for (size_type i = 0; i != matrix.size1(); ++ i) {
 	if (i)
 	  karma::generate(iter, karma::lit(", "));
-	karma::generate(iter, '(' << (real % ", ") << ')',
+	
+	karma::generate(iter, '(' << (('B' << base64) % ", ") << ')',
 			boost::make_iterator_range(matrix.begin(i), matrix.end(i)));
       }
       karma::generate(iter, karma::lit(')'));
     }
     karma::generate(iter, karma::lit('\n'));
-#endif
-#if 1
-    const matrix_type& matrix = posterior.matrix;
-    
-    if (matrix.empty())
-      os << '\n';
-    else {
-      os << '(';
-      for (size_type i = 0; i != matrix.size1(); ++ i) {
-	if (i)
-	  os << ", ";
-	os << '(';
-	matrix_type::const_iterator iter_begin = matrix.begin(i);
-	matrix_type::const_iterator iter_end   = matrix.end(i);
-	if (iter_begin != iter_end) {
-	  std::copy(iter_begin, iter_end - 1, std::ostream_iterator<double>(os, ", "));
-	  os << *(iter_end - 1);
-	}
-	os << ')';
-      }
-      os << ')' << '\n';
+
+    posterior.output = std::string(buffer.begin(), buffer.end());
+  }
+};
+
+struct PosteriorReducer : public PosteriorMapReduce
+{
+  typedef std::pair<posterior_type, queue_reducer_type*> posterior_queue_type;
+  typedef std::vector<posterior_queue_type, std::allocator<posterior_queue_type> > posterior_queue_set_type;
+
+  struct heap_compare_type
+  {
+    bool operator()(const posterior_queue_type* x, const posterior_queue_type* y) const
+    {
+      return x->first > y->first;
     }
-#endif
+  };
+  
+  typedef std::vector<posterior_queue_type*, std::allocator<posterior_queue_type*> > heap_base_type;
+  typedef std::priority_queue<posterior_queue_type*, heap_base_type, heap_compare_type> heap_type;
+
+  typedef std::vector<posterior_queue_type*, std::allocator<posterior_queue_type*> > queue_type;
+
+  typedef boost::shared_ptr<std::ostream> ostream_ptr_type;
+  
+  ostream_ptr_type os;
+  queue_reducer_set_type& queues;
+  bool flush_;
+  
+  PosteriorReducer(const path_type& path, queue_reducer_set_type& __queues) : os(), queues(__queues), flush_(false)
+  {
+    if (! path.empty()) {
+      flush_ = (path == "-"
+		|| (boost::filesystem::exists(path)
+		    && ! boost::filesystem::is_regular_file(path)));
+      
+      os.reset(new utils::compress_ostream(path, 1024 * 1024));
+      os->precision(20);
+    }
+  }
+  
+  void operator()() throw()
+  {
+    std::ostream* stream = os.get();
+
+    posterior_queue_set_type posteriors(queues.size());
+    
+    heap_type heap;
+    queue_type queue;
+    queue_type queue_next;
+
+    size_type id = 0;
+    
+    for (size_type shard = 0; shard != queues.size(); ++ shard) {
+      posteriors[shard].second = &queues[shard];
+      
+      queue.push_back(&posteriors[shard]);
+    }
+    
+    int non_found_iter = 0;
+    while (! heap.empty() || ! queue.empty()) {
+      bool found = false;
+
+      if (! heap.empty() && heap.top()->first.id == id) {
+	posterior_queue_type* posterior_queue = heap.top();
+	heap.pop();
+	
+	++ id;
+	
+	if (stream) {
+	  *stream << posterior_queue->first.output;
+	  
+	  if (flush_)
+	    *stream << std::flush;
+	}
+	
+	queue.push_back(posterior_queue);
+	
+	found = true;
+      }
+
+      if (! queue.empty()) {
+	for (size_type i = 0; i != queue.size(); ++ i) {
+	  if (queue[i]->second->pop_swap(queue[i]->first, true)) {
+	    if (queue[i]->first.id != size_type(-1))
+	      heap.push(queue[i]);
+	    
+	    found = true;
+	  } else
+	    queue_next.push_back(queue[i]);
+	}
+	
+	queue.swap(queue_next);
+	queue_next.clear();
+      }
+      
+      non_found_iter = loop_sleep(found, non_found_iter);
+    }
+  }
+  
+  int loop_sleep(bool found, int non_found_iter)
+  {
+    if (! found) {
+      boost::thread::yield();
+      ++ non_found_iter;
+    } else
+      non_found_iter = 0;
+  
+    if (non_found_iter >= 50) {
+      struct timespec tm;
+      tm.tv_sec = 0;
+      tm.tv_nsec = 2000001;
+      nanosleep(&tm, NULL);
+    
+      non_found_iter = 0;
+    }
+    return non_found_iter;
   }
 };
 
@@ -1561,15 +1833,16 @@ void posterior(const ttable_type& ttable_source_target,
   
   typedef reducer_type::bitext_type    bitext_type;
   typedef reducer_type::posterior_type posterior_type;
-  
-  typedef reducer_type::queue_mapper_type  queue_mapper_type;
-  typedef reducer_type::queue_reducer_type queue_reducer_type;
+
+  typedef reducer_type::queue_mapper_type      queue_mapper_type;
+  typedef reducer_type::queue_reducer_type     queue_reducer_type;
+  typedef reducer_type::queue_reducer_set_type queue_reducer_set_type;
   
   typedef std::vector<mapper_type, std::allocator<mapper_type> > mapper_set_type;
   
-  queue_mapper_type  queue(threads * 4096);
-  queue_reducer_type queue_source_target;
-  queue_reducer_type queue_target_source;
+  queue_mapper_type  queue(threads * 1024);
+  queue_reducer_set_type queue_source_target(threads, queue_reducer_type(1024));
+  queue_reducer_set_type queue_target_source(threads, queue_reducer_type(1024));
   
   boost::thread_group reducer;
   reducer.add_thread(new boost::thread(reducer_type(posterior_source_target_file, queue_source_target)));
@@ -1579,8 +1852,8 @@ void posterior(const ttable_type& ttable_source_target,
   for (int i = 0; i != threads; ++ i)
     mapper.add_thread(new boost::thread(mapper_type(Infer(ttable_source_target, ttable_target_source),
 						    queue,
-						    queue_source_target,
-						    queue_target_source)));
+						    queue_source_target[i],
+						    queue_target_source[i])));
     
   bitext_type bitext;
   bitext.id = 0;
@@ -1621,10 +1894,6 @@ void posterior(const ttable_type& ttable_source_target,
     queue.push_swap(bitext);
   }
   mapper.join_all();
-  
-  queue_source_target.push(posterior_type());
-  queue_target_source.push(posterior_type());
-  
   reducer.join_all();
 
   utils::resource posterior_end;
