@@ -18,7 +18,6 @@
 #include <cmath>
 #include <climits>
 
-#include <deque>
 #include <memory>
 
 #include "cicada_nn_ngram_impl.hpp"
@@ -30,6 +29,7 @@
 #include "utils/random_seed.hpp"
 #include "utils/compress_stream.hpp"
 #include "utils/resource.hpp"
+#include "utils/chunk_vector.hpp"
 
 #include <boost/algorithm/string/trim.hpp>
 #include <boost/random.hpp>
@@ -208,7 +208,7 @@ struct TaskAccumulate
   typedef utils::lockfree_list_queue<gradient_type*, std::allocator<gradient_type*> > queue_merger_type;
   typedef std::vector<queue_merger_type, std::allocator<queue_merger_type> > queue_merger_set_type;
   
-  typedef std::deque<gradient_type, std::allocator<gradient_type> > gradient_set_type;
+  typedef utils::chunk_vector<gradient_type, 1024 / sizeof(gradient_type), std::allocator<gradient_type> > gradient_set_type;
   
   TaskAccumulate(const Learner& learner,
 		 const data_type& data,
@@ -260,7 +260,7 @@ struct TaskAccumulate
 	  found = true;
 	}
       
-      if (! learn_finished && mapper_.pop(batch, true)) {
+      if (! learn_finished && ((grad = allocate()) != 0) && mapper_.pop(batch, true)) {
 	
 	found = true;
 	
@@ -271,19 +271,6 @@ struct TaskAccumulate
 	  
 	  learn_finished = true;
 	} else {
-	  gradient_type* grad = 0;
-	  
-	  for (size_type j = 0; j != gradients_.size(); ++ j)
-	    if (gradients_[j].shared() == shard_size) {
-	      grad = &gradients_[j];
-	      break;
-	    }
-	  
-	  if (! grad) {
-	    gradients_.push_back(gradient_type(theta_.dimension_embedding_, theta_.dimension_hidden_, theta_.order_));
-	    grad = &gradients_.back();
-	  }
-	  
 	  grad->clear();
 	  
 	  // use of batch from batch * batch_size and (batch + 1) * batch_size
@@ -309,6 +296,24 @@ struct TaskAccumulate
     // rescale current model
     theta_.rescale();
   }
+
+  gradient_type* allocate()
+  {
+    gradient_type* grad = 0;
+    
+    for (size_type j = 0; j != gradients_.size(); ++ j)
+      if (gradients_[j].shared() >= mergers_.size() || gradients_[j].shared() == 0) {
+	grad = &gradients_[j];
+	break;
+      }
+    
+    if (! grad && gradients_.size() < 128) {
+      gradients_.push_back(gradient_type(theta_.dimension_embedding_, theta_.dimension_hidden_, theta_.order_));
+      grad = &gradients_.back();
+    }
+    
+    return grad;
+  }
   
   inline
   int loop_sleep(bool found, int non_found_iter)
@@ -455,14 +460,16 @@ void learn_online(const Learner& learner,
     
     if (debug)
       std::cerr << "log-likelihood: " << static_cast<double>(log_likelihood) << std::endl
-		<< "perplexity: " << std::exp(- static_cast<double>(log_likelihood)) << std::endl;
+		<< "entropy: "    << std::exp(- static_cast<double>(log_likelihood)) << std::endl
+		<< "perplexity: " << (- static_cast<double>(log_likelihood) / std::log(2.0)) << std::endl;
     
     if (debug)
       std::cerr << "cpu time:    " << end.cpu_time() - start.cpu_time() << std::endl
 		<< "user time:   " << end.user_time() - start.user_time() << std::endl;
     
     // shuffle ngrams!
-    std::random_shuffle(batches.begin(), batches.end());
+    boost::random_number_generator<boost::mt19937> gen(tasks.front().generator_);
+    std::random_shuffle(batches.begin(), batches.end(), gen);
     
     if (mix_average) {
       for (size_type i = 1; i != tasks.size(); ++ i)
