@@ -483,8 +483,10 @@ struct PYPGraph
     for (size_type t = 1; t != T; ++ t) {
       for (id_type prev = 0; prev != K; ++ prev)
 	for (id_type next = 1; next != K; ++ next)
-	  if (model.cache_transition(prev, next) > cutoff[t])
-	    alpha(t, next) += alpha(t - 1, prev) * model.cache_transition(prev, next) * model.cache_emission(next, sentence[t - 1]);
+	  if (model.cache_transition(prev, next) > cutoff[t]) {
+	    //alpha(t, next) += alpha(t - 1, prev) * model.cache_transition(prev, next) * model.cache_emission(next, sentence[t - 1]);
+	    alpha(t, next) += alpha(t - 1, prev) * model.cache_emission(next, sentence[t - 1]);
+	  }
       
       double scale = std::accumulate(alpha.begin(t), alpha.end(t), 0.0);
       scale = (scale == 0.0 ? 1.0 : scale);
@@ -522,7 +524,8 @@ struct PYPGraph
       positions.clear();
       for (id_type prev = 1; prev != K; ++ prev) 
 	if (alpha(t - 1, prev) != 0.0 && model.cache_transition(prev, state) > cutoff[t]) {
-	  probs.push_back(alpha(t - 1, prev) * model.cache_transition(prev, state) * model.cache_emission(state, sentence[t - 1]));
+	  //probs.push_back(alpha(t - 1, prev) * model.cache_transition(prev, state) * model.cache_emission(state, sentence[t - 1]));
+	  probs.push_back(alpha(t - 1, prev) * model.cache_emission(state, sentence[t - 1]));
 	  positions.push_back(prev);
 	}
       
@@ -1087,7 +1090,7 @@ int main(int argc, char ** argv)
       // sample other parameters...
       if (debug >= 3)
 	std::cerr << "sample parameters" << std::endl;
-	    
+      
       model.sample_parameters(sampler, resample_iterations);
       
       if (debug >= 2)
