@@ -935,13 +935,15 @@ void learn_online(const Learner& learner,
 
     // shuffle bitexts!
     {
+      boost::random_number_generator<boost::mt19937> gen(tasks.front().generator_);
+      
       typename batch_set_type::iterator biter     = batches.begin();
       typename batch_set_type::iterator biter_end = batches.end();
       
       while (biter < biter_end) {
 	typename batch_set_type::iterator iter_end = std::min(biter + utils::bithack::max(4096 / batch_size, 1), biter_end);
 	
-	std::random_shuffle(biter, iter_end);
+	std::random_shuffle(biter, iter_end, gen);
 	biter = iter_end;
       }
     }
